#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/common.h>
#define F_CPU 1000000UL //actually we will not use this, I guess :) (and it should better be 8000000UL, but that'll be a guess :)  )
#include <util/delay.h> //and this, I suppose :)
#include <limits.h>

#define TICK_DURATION_US 10

#define timetype unsigned int

#if (defined(__AVR_ATtiny13__))
  #define ENABLE_TIMER_INTERRUPT( )       ( TIMSK0 |= ( 1<< OCIE0A ) )
  #define DISABLE_TIMER_INTERRUPT( )      ( TIMSK0 &= ~( 1<< OCIE0A ) )
  #define CLEAR_TIMER_INTERRUPT( )        ( TIFR0 |= ((1 << OCF0A) ) )
  #define ENABLE_EXTERNAL0_INTERRUPT( )   ( GIMSK |= ( 1 << INT0 ) )
  #define DISABLE_EXTERNAL0_INTERRUPT( )  ( GIMSK &= ~( 1 << INT0 ) )
  
  #define OUT_HIGH( )                     ( PORTB |=  ( 1 << PB2  ) )
  #define OUT_LOW( )                      ( PORTB &= ~( 1 << PB2  ) )
  #define ENABLE_OUTPUT( )                ( DDRB  =   ( 1 << DDB2 ) )
  
  #define RX_PIN           PB1
  #define TCCR             TCCR0A             //!< Timer/Counter Control Register
  #define TCCR_P           TCCR0B             //!< Timer/Counter Control (Prescaler) Register
  #define OCR              OCR0A              //!< Output Compare Register
  #define EXT_IFR          GIFR               //!< External Interrupt Flag Register
  #define EXT_ICR          MCUCR              //!< External Interrupt Control Register

#else
  #warning Selected AVR device is not supported
#endif

#define STATE_SIGNAL_LOST -1
#define STATE_IDLE 0
#define STATE_ENABLE1 1
#define STATE_ENABLE2 2

#define MAX_WAVE_PERIOD 200
#define MIN_WAVE_PERIOD 1

const timetype controlIgnoreMin =  ((timetype)   300) / TICK_DURATION_US;
const timetype controlIgnoreMax =  ((timetype)  4000) / TICK_DURATION_US;
const timetype controlIdle      =  ((timetype)  1400) / TICK_DURATION_US;
const timetype controlE1        =  ((timetype)  2000) / TICK_DURATION_US;


int state = STATE_IDLE;
timetype wavePeriod = MIN_WAVE_PERIOD;
timetype lastWaveChange = 0;

//timers
timetype time = 0;
timetype lastSignalCheckTime = 0, lastControlTime = 0, lastHighEdge = 0;

timetype timeDiff(timetype time1, timetype time2) {
    if (time1 == time2) return 0;
    if (time1 < time2) {
        return time2-time1;
    } else {
        return (UINT_MAX-time1)+time2;
    }
}

int outState = 0;
timetype buzzTime = 0;
//makes the output oscilate with period equals 2*wavePeriod
void buzz() {
    if (timeDiff(buzzTime, time) >= wavePeriod) {
        if (outState == 0) {
			outState = 1;
			OUT_HIGH( );
		} else {
			outState = 0;
			OUT_LOW( );
		}
        buzzTime = time;
    }
}

ISR(INT0_vect)
{
    //we are here on any change on RX_PIN
    if (((PINB >> RX_PIN) & 1) > 0) {
        lastHighEdge = time;
    } else {
        timetype lastSignalDuration = timeDiff(lastHighEdge, time);
		if (lastSignalDuration < controlIgnoreMin || lastSignalDuration > controlIgnoreMax) {
			//ignore this signal. seems invalid
        } else if (lastSignalDuration <= controlIdle) {
            state = STATE_IDLE;
			lastControlTime = 0;
        } else if (lastSignalDuration > controlIdle && lastSignalDuration <= controlE1) {
            state = STATE_ENABLE1;
			lastControlTime = 0;
        } else if (lastSignalDuration > controlE1) {
            state = STATE_ENABLE2;
			lastControlTime = 0;
        }
    }
}

ISR(TIM0_COMPA_vect) {
    //we are here every TICK_DURATION_US microseconds
    time++;
}

void initInterrupts() {
    //PORT
    //PORTB = ( 1 << RX_PIN );        // RX_PIN is input, tri-stated.
    ENABLE_OUTPUT( );
    
    //if we suppose that ATtiny works at 1MHz by default 
    //and we disable main clock prescaler which was 8
    //and set timer prescaler to 8
    //than we can suppose that timer can generate roughly 1 tick per microseconds
    OCR = TICK_DURATION_US;
    
    // Timer0
    DISABLE_TIMER_INTERRUPT( );
    TCCR = 0x00;    //Init.
    TCCR_P = 0x00;    //Init.
    TCCR |= (1 << WGM01);            // Timer in CTC mode.
    TCCR_P |=  ( 1 << CS01 );        // Divide by 8 prescaler.
    ENABLE_TIMER_INTERRUPT( );

    //External interrupt
    DISABLE_EXTERNAL0_INTERRUPT( );
    EXT_ICR = 0x00;                   // Init.
    EXT_ICR |= ( 1 << ISC00 );        // Interrupt sense control: any change
    ENABLE_EXTERNAL0_INTERRUPT( );    // Turn external interrupt on.
  
    sei();
}

int main(void) {
	CLKPR = (1 << CLKPCE); //enable main clock control
	CLKPR = 0x00; //disable main clock prescaler (was 8, now 1)
	//OSCCAL  = 0x7F; //multiply clock speed by 2
	
    initInterrupts();
    
    while (1) {
        if (state == STATE_SIGNAL_LOST || state == STATE_ENABLE2) {
            if (timeDiff(lastWaveChange, time) > 110) {
                wavePeriod--;
                if (wavePeriod < MIN_WAVE_PERIOD) {
                    wavePeriod = MAX_WAVE_PERIOD;
                }
                lastWaveChange = time;
            }
            
            buzz();
        } else if (state == STATE_ENABLE1) {
			if (timeDiff(lastWaveChange, time) > 90) { //Star Wars blasters :)
                wavePeriod++;
                if (wavePeriod > MAX_WAVE_PERIOD) {
                    wavePeriod = MIN_WAVE_PERIOD;
                }
                lastWaveChange = time;
            }
            
            buzz();
        } else if (outState != 0) {
            OUT_LOW( ); //no sound played. disable speakers for lower power consumption
        }
        
        //check if the signal is lost (no valid control impulse for a long time)
        if (state != STATE_SIGNAL_LOST && lastSignalCheckTime != time) {
			lastControlTime++;
			if (lastControlTime == UINT_MAX) {
				state = STATE_SIGNAL_LOST;
			}
			lastSignalCheckTime = time;
        }
    }
    return 0;
}